#include <boost/mpl/assert_is_same.hpp>
#include <boost/mpl/bind.hpp>
#include <boost/mpl/max_element.hpp>
#include <boost/mpl/vector.hpp>

#include "boost/type_traits/is_better_conversion.hpp"

namespace mpl = boost::mpl;
using namespace mpl::placeholder;

// algorithm
template <typename From, typename ToSequence>
struct best_conversion
{
private:
    typedef mpl::bind3< // is_better_conversion< T,_2,_1 >
		  boost::is_better_conversion<_,_,_>
        , From
        , mpl::_2
        , mpl::_1
        > conversion_compare;

    typedef typename mpl::max_element<
          ToSequence
//        , conversion_compare
		, boost::is_better_conversion<From, mpl::_2, mpl::_1>
        >::type max_it;

public:
    typedef typename max_it::type
        type;
};

// usage
int main()
{
	// This one is easy...
    typedef best_conversion<
          int
        , mpl::vector< void*, float, int, char >
        >::type result1;

	// ...of course |int| converts best to an |int|:
    BOOST_MPL_ASSERT_IS_SAME(result1, int);

	// OK, this one is harder, but |float| seems better than |char|...:
    typedef best_conversion<
          int
        , mpl::vector< void*, float, char >
        >::type result2;

	// ...so it's all good:
    BOOST_MPL_ASSERT_IS_SAME(result2, float);

	// But just change the order of the types in the list...
	typedef best_conversion<
          int
        , mpl::vector< void*, char, float >
        >::type result3;

	// ...and voila! it's broken!
    //BOOST_MPL_ASSERT_IS_SAME(result3, float);
}
